#######################################################################################################################################################################################################
#
# Changes to the method
# General
#     2022-Feb-01: add method for LogisticVC
#     2022-Feb-01: add method for PowerVC
#     2022-Feb-01: add method for WeibullVC
#     2022-Feb-01: add method for ComplexVC
#     2022-Feb-01: add documentation
#     2022-Jul-08: deflate documentations
#
#######################################################################################################################################################################################################
"""

    relative_hydraulic_conductance(vc::ComplexVC{FT}, p_25::FT) where {FT<:AbstractFloat}
    relative_hydraulic_conductance(vc::LogisticVC{FT}, p_25::FT) where {FT<:AbstractFloat}

Return the hydraulic conductance ralative to maximum at reference temperature, given
- `vc` `ComplexVC`, `LogisticVC`, `PowerVC`, or `WeibullVC` type vulnerability curve
- `p_25` Equivalent xylem water pressure at 298.15 K in `[MPa]` (surface tension correction made)
"""
relative_hydraulic_conductance(vc::ComplexVC{FT}, p_25::FT) where {FT<:AbstractFloat} = (
    @unpack PS, VCS = vc;
    @assert sum(PS) ≈ 1 "Probabilities of VCs must sum up to 1!";
    @assert length(PS) == length(VCS) "Lengths of VC curves and probabilities must equal!";

    _kr::FT = 0;
    for _i in eachindex(PS)
        _kr += relative_hydraulic_conductance(VCS[_i], p_25) * PS[_i];
    end;

    return _kr
);

relative_hydraulic_conductance(vc::LogisticVC{FT}, p_25::FT) where {FT<:AbstractFloat} = (
    if p_25 >= 0
        return FT(1)
    end;

    @unpack A, B = vc;

    return max( eps(FT), (1 - 1/(1 + A * exp(B * p_25))) * (A+1)/A )
);

relative_hydraulic_conductance(vc::PowerVC{FT}, p_25::FT) where {FT<:AbstractFloat} = (
    if p_25>=0
        return FT(1)
    end;

    @unpack A, B = vc;

    return max( eps(FT), 1 / (1 + A*(-p_25)^B) )
);

relative_hydraulic_conductance(vc::WeibullVC{FT}, p_25::FT) where {FT<:AbstractFloat} = (
    if p_25>=0
        return FT(1)
    end;

    @unpack B, C = vc;

    return max( eps(FT), exp( -1 * (-p_25 / B) ^ C ) )
);
